# Duo, A crazy eights game.
# Copyright (C) 2009 The Duo developers.
#
# This file is part of Duo.
#
# Duo is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

"""Game scheduler and arbiter.
"""

import logging
import sys

from duo import protocol
from duo import game


class GameAlreadyRunningError(Exception):
    """Error raised on multiple game start."""
    pass

class InvalidMoveError(Exception):
    """Error raised on invalid move."""
    pass


# TODO: needs refactoring.
class PlayerInfo(object):
    def __init__(self, name):
        self.name = name
        self.hand = []
        self.actions = []
        self.ready = False
        self.drawn = False


class Carpet(object):
    def __init__(self):
        self.current_color = None
        self.deck = None
        self.heap = None
        self.kitty = None

    def reset(self):
        self.deck = game.Deck()
        self.heap = game.Deck()
        self.kitty = game.Kitty()


class Action(object):
    def __init__(self, player):
        self.player = player

    def __hash__(self):
        result = hash(self.__class__)
        for o in self.__dict__.keys():
            result = (result + hash(o)) & sys.maxint
        return result

    def __eq__(self, action):
        if action is None:
            return False
        else:
            return self.__class__ == action.__class__ and self.player == action.player

    def __ne__(self, action):
        return not self.__eq__(action)

    def __repr__(self):
        return self.__str__()

    def __str__(self):
        return self.__class__.__name__ + ': ' + str(self.player)


class CardAction(Action):
    def __init__(self, player, card, color=None):
        Action.__init__(self, player)
        self.card = card
        self.color = color if color else card.color

    def __eq__(self, action):
        return Action.__eq__(self, action) and self.card == action.card and self.color == action.color

    def __str__(self):
        return '(' + self.__class__.__name__ + str(self.card) + ',' + self.color + ')'


class DrawAction(Action):
    def __init__(self, player):
        Action.__init__(self, player)


class PassAction(Action):
    def __init__(self, player):
        Action.__init__(self, player)


class ReadyAction(Action):
    """Action emitted by players when they are ready to start.
    """
    def __init__(self, player):
        Action.__init__(self, player)


class PlayersRing(list):
    """The players list considering their order and their hands.

    Every entry of this list is a tuple containing first the player name
    and then this player hand and its status.
    """

    # Game directions
    CW = +1
    CCW = -1

    def __init__(self):
        list.__init__(self)
        self._direction = PlayersRing.CW

        # Current player index is negative until the first player is elected.
        self._current = -1
        self._dealer = -1

    def add(self, player):
        """Add a new player."""
        if player in [pinfo.name for pinfo in self]:
            raise protocol.PlayerAlreadyAuthenticatedError()
        self.append(PlayerInfo(player))

    def remove(self, player):
        """Remove a player."""
        for pinfo in self:
            if pinfo.name == player:
                self.remove(pinfo)
                return

    def get_list(self):
        """Return the player names."""
        return [pinfo.name for pinfo in self]

    def get_list_except(self, player):
        """Retrieve the list of player, minus the one given."""
        return [pinfo.name for pinfo in self if pinfo.name != player]

    def update_player_actions(self, player, actions):
        """Update the possible actions for the given player."""
        for pinfo in self:
            if pinfo.name == player:
                pinfo.actions = actions
                break

    def update_player_status(self, player, ready=None, drawn=None):
        """Update the status of the given player."""
        for pinfo in self:
            if pinfo.name == player:
                if not ready is None:
                    pinfo.ready = ready
                if not drawn is None:
                    pinfo.drawn = drawn
                break

    def reset(self):
        """Reset the status of all players."""
        self._dealer = (self._dealer + 1) % len(self)
        self._current = -1
        self._direction = PlayersRing.CW
        for pinfo in self:
            pinfo.ready = False
            pinfo.drawn = False
            pinfo.hand = []

    def get_player_hand(self, player):
        """Retrieve the hand of the given player."""
        entries = [pinfo.hand for pinfo in self if pinfo.name == player]
        if entries:
            return entries[0]
        else:
            return None

    def get_player_actions(self, player):
        """Retrieve the actions of the given player."""
        entries = [pinfo.actions for pinfo in self if pinfo.name == player]
        if entries:
            return entries[0]
        else:
            return None

    def has_player_drawn(self, player):
        """Check if the given player has already drawn a card."""
        entries = [pinfo.drawn for pinfo in self if pinfo.name == player]
        if entries:
            return entries[0]
        else:
            return None

    def current(self):
        """Return the current player."""
        return self[self._current].name if self._current >= 0 else None

    def next(self):
        """Return the next player to play."""
        return self[self._next_current()].name

    def change_direction(self):
        """Change the direction of progression in the players ring."""
        self._direction = - self._direction

    def pass_turn(self):
        """Remember to pass for next player turn computation."""
        self._direction = 2 * self._direction

    def get_deal_order(self):
        """Return the ordered indexes of the players for the next game.

        Order is determined by the dealer position. The first player
        of the list if the first on the left of the dealer. The rest
        of the list is build with the rest of the players, clockwise.
        """
        deal_order = []
        for i in xrange(len(self)):
            deal_order.append((i + 1 + self._dealer) % len(self))
        return deal_order

    def compute_next(self):
        """Compute a new TurnMessage with the next player to play."""
        # TODO: This should be written as a referee rule.
        # Reset the drawing status of the current player.
        self[self._current].drawn = False

        self._current = self._next_current()
        self._direction = self._direction / abs(self._direction)
        return protocol.TurnMessage(self.current())

    def is_ready(self, player=None):
        """Check readiness of players.

        If player is set, then return True if this player exists and
        is ready, False otherwise. If player is None, return True if
        all players are ready, false otherwise.
        """
        if player:
            entries = [pinfo.ready for pinfo in self if pinfo.name == player]
            if entries:
                return entries[0]
            return False
        else:
            for pinfo in self:
                if pinfo.ready == False:
                    return False
            return True

    def _next_current(self):
        """Return the next current pointer value."""
        if self._current >= 0:
            return (self._current + self._direction) % len(self)
        else:
            # Compute initial player...
            return (self._dealer + self._direction) % len(self)


class RuleSet(set):
    """
    """
    def __init__(self, referee, options):
        set.__init__(self)
        self._referee = referee

        ruleset = options['ruleset']
        m = __import__('duo.rules.' + ruleset, fromlist=ruleset)
        self.update(m.init(options))

    def get_options(self, man):
        """Compute what the referee whould do now.
        """
        options = []
        for rule in self:
            options.extend(rule.check(
                man, self._referee.get_players(), self._referee.get_carpet()))
        return options

    def update_actions(self, bundles):
        """Compute the possible actions for the players
        """
        ring = self._referee.get_players()
        for player in ring.get_list():
            options = self.get_options(player)
            ring.update_player_actions(player, options)
            bundles[player].append(protocol.HandMessage(
                ring.get_player_hand(player), set((o[0] for o in options))))


#REFEREE_OPTIONS = {
#    'game': ('stupid-default-name', 'The name to use when opening the game'),
#    'cards': (7, 'The number of cards to deal'),
#    'ruleset': ('DEFAULT_RULESET', 'The ruleset to use')
#}
class Referee(protocol.ProtocolAdapter):
    """The game referee.
    """
    WAITING = 'waiting'
    PLAYING = 'playing'

    def __init__(self, transport_layer, transport_options, referee_options):
        protocol.ProtocolAdapter.__init__(self, 'referee')
        self.connect(transport_layer, transport_options)

        self.register(protocol.StartMessage, self.on_start_cb, None)
        self.register(protocol.ActionMessage, self.on_action_cb, None)

        self._players = PlayersRing()
        self._carpet = Carpet()
        self._options = referee_options
        self._ruleset = RuleSet(self, referee_options)

        self._game_status = Referee.WAITING

        logging.getLogger('duo.referee').info('Referee ready.')

    def leave(self):
        """Leave and finish the game"""
        # Escape twice to leave the game, then the transport layer.
        self._endpoint.escape()
        self._endpoint.escape()

    def send_except(self, message, player):
        """Send the given message to everyone except the given player."""
        for player_ in self._players.get_list_except(player):
            self.send_to(message, player_)

    def on_endpoint_join(self, player, emitter):
        """Callback invoked by the transport layer on new connection."""
        if self._game_status != Referee.WAITING:
            raise GameAlreadyRunningError()
        self._players.add(player)
        logging.getLogger('duo.referee').info('%s joined the game', player)
        self.broadcast(protocol.AuthMessage(self._players.get_list()))

    def on_endpoint_escape(self, player, emitter):
        """Callback invoked by the transport layer on disconnection."""
        # TODO: if game is running, should probably be interrupted
        self._players.remove(player)
        logging.getLogger('duo.referee').info('%s left the game', player)

    def on_start_cb(self, key, player, args):
        """Callback invoked on reception of a Start message."""
        if key == self._options['master_key']:
            logging.getLogger('duo.referee').info('Starting game')

            self._game_status = Referee.PLAYING
            self._players.reset()
            self._carpet.reset()

            options = self._ruleset.get_options(None)
            print '? on_start_cb:', options

            self._play(None, None)
        else:
            logging.getLogger('duo.referee').warning(
                'received start with wrong key')

    def on_action_cb(self, message, player, args):
        """Callback invoked on reception of an action message."""
        actions = [pair for pair in self._players.get_player_actions(player)
                   if pair[0] == message]
        if not len(actions):
            raise InvalidMoveError('Invalid move. Are you a cheater ?')

        # TODO: Add the Big Lock
        self._play(player, actions[0][1])

    def get_players(self):
        """Return the players list."""
        return self._players

    def get_carpet(self):
        """Return the currently asked color."""
        return self._carpet

    def _play(self, player, effect):
        """Check and apply - or reject - a move."""
        try:
            bc_bundle = protocol.MessageBundle()
            bundles = {}
            for p in self._players.get_list():
                bundles[p] = protocol.MessageBundle()

            # Since the referee accepts a move by sending a new heap,
            # and the UI uses this message to represent the move, the
            # HeapMessage must be sent first.
            # Note below that the HeapMessage, if any, is in bc_bundle.

            if effect:
                effect.apply(self.get_players(), self.get_carpet(), bundles, bc_bundle)

            # Once a move is validated and its effects applied, the
            # referee and the players get their new actions. The
            # referee applies immediately any action it can undertake.
            options = self._ruleset.get_options(None)
            for action, effect in options:
                effect.apply(self.get_players(), self.get_carpet(), bundles, bc_bundle)

            self._ruleset.update_actions(bundles)

            if len(bc_bundle):
                print '? _play: all', bc_bundle
                self.broadcast(bc_bundle)

            for p in bundles.keys():
                print '? _play:', p, bundles[p]
                self.send_to(bundles[p], p)
            print '----------'
        except Exception, ex:
            self.send_to(protocol.ErrorMessage(ex), player)
