# Duo, A crazy eights game.
# Copyright (C) 2004,2009 The Duo developers.
#
# This file is part of Duo.
#
# Duo is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

"""Official set of rules.
"""

from duo import protocol
from duo import game
from duo.referee import CardAction, DrawAction, PassAction, ReadyAction
from duo.rules import Rule, Effect, no_referee, only_referee, no_kitty, at_turn




class CardEffect(Effect):
    def __init__(self, player, card, color=None):
        Effect.__init__(self, player)
        self.card = card
        self.color = color if color else card.color

    def apply(self, ring, carpet, bundles, bc_bundle):
        carpet.heap.append(self.card)
        carpet.current_color = self.color
        bc_bundle.append(protocol.HeapMessage(
            carpet.heap.peek(), carpet.current_color, self.player))
        if not self.player is None:
            ring.get_player_hand(self.player).remove(self.card)
        bc_bundle.append(ring.compute_next())


class ReverseCardEffect(CardEffect):
    def __init__(self, player, card, color=None):
        CardEffect.__init__(self, player, card, color)

    def apply(self, ring, carpet, bundles, bc_bundle):
        # in case of a two-player game, 'R' card behavior is equivalent to 'P'
        if len(ring) == 2:
            ring.pass_turn()
        else:
            ring.change_direction()
        bc_bundle.append(protocol.ReverseMessage())
        CardEffect.apply(self, ring, carpet, bundles, bc_bundle)


class PassCardEffect(CardEffect):
    def __init__(self, player, card, color=None):
        CardEffect.__init__(self, player, card, color)

    def apply(self, ring, carpet, bundles, bc_bundle):
        ring.pass_turn()
        CardEffect.apply(self, ring, carpet, bundles, bc_bundle)


class DrawTwoCardEffect(CardEffect):
    def __init__(self, player, card, color=None):
        CardEffect.__init__(self, player, card, color)

    def apply(self, ring, carpet, bundles, bc_bundle):
        carpet.kitty.type = 'D'
        for times in range(2):
            carpet.kitty.append(carpet.deck.pop())
        CardEffect.apply(self, ring, carpet, bundles, bc_bundle)


class DrawFourCardEffect(CardEffect):
    def __init__(self, player, card, color):
        CardEffect.__init__(self, player, card, color)

    def apply(self, ring, carpet, bundles, bc_bundle):
        carpet.kitty.type = 'X'
        for times in range(4):
            carpet.kitty.append(carpet.deck.pop())
        CardEffect.apply(self, ring, carpet, bundles, bc_bundle)


class DrawEffect(Effect):
    def __init__(self, player):
        Effect.__init__(self, player)

    def apply(self, ring, carpet, bundles, bc_bundle):
        if len(carpet.kitty):
            ring.get_player_hand(self.player).extend(carpet.kitty.pick_up())
        else:
            ring.get_player_hand(self.player).append(carpet.deck.pop())

        ring.update_player_status(self.player, drawn=True)


class PassEffect(Effect):
    def __init__(self, player):
        Effect.__init__(self, player)

    def apply(self, ring, carpet, bundles, bc_bundle):
        if carpet.kitty.taken:
            carpet.kitty = game.Kitty()
        if ring.get_player_hand(self.player) != []:
            bc_bundle.append(ring.compute_next())


class ReadyEffect(Effect):
    def __init__(self, player):
        Effect.__init__(self, player)

    def apply(self, ring, carpet, bundles, bc_bundle):
        ring.update_player_status(self.player, ready=True)


class FirstCardEffect(Effect):
    def __init__(self, player):
        Effect.__init__(self, player)

    def apply(self, ring, carpet, bundles, bc_bundle):
        # Deal the cards
        card = carpet.deck.pop()
        while card.color == 'N':
            bc_bundle.append(protocol.HeapMessage(card, card.color, None))
            carpet.heap.append(card)
            card = carpet.deck.pop()

        Effects[card.type](None, card).apply(ring, carpet, bundles, bc_bundle)


class DealEffect(Effect):
    def __init__(self, player, cards):
        Effect.__init__(self, player)
        self._cards = cards

    def apply(self, ring, carpet, bundles, bc_bundle):
        carpet.deck.generate()
        carpet.deck.shuffle()

        # Deal the cards
        hands = [[] for i in range(len(ring))]
        for times in range(self._cards):
            for i in range(len(hands)):
                hands[i].append(carpet.deck.pop())

        for i in ring.get_deal_order():
            play = ring[i].name
            ring[i].hand.extend(hands[i])


class SwapStacksEffect(Effect):
    def __init__(self):
        Effect.__init__(self, None)

    def apply(self, ring, carpet, bundles, bc_bundle):
        # Keep only the last card on the heap.
        card = carpet.heap.pop()
        carpet.deck = carpet.heap
        carpet.heap = game.Deck()
        carpet.heap.append(card)

        carpet.deck.shuffle()


class EndOfGameEffect(Effect):
    def __init__(self, winner):
        Effect.__init__(self, winner)

    def apply(self, ring, carpet, bundles, bc_bundle):
        if len(carpet.kitty):
            print 'Last Kitty!', ring.current(), ring.get_player_hand(ring.current())
            ring.get_player_hand(ring.current()).extend(carpet.kitty.pick_up())
        bc_bundle.append(protocol.WinnerMessage(self.player))


class ColorRule(Rule):
    """Rule allowing to play a card with the same color."""
    def __init__(self):
        Rule.__init__(self)

    @no_referee
    @no_kitty
    @at_turn
    def check(self, player, ring, carpet):
        """Check this rule.

        This rule should match only for cards with color matching the
        heap top card color. It always fails if a kitty is present or
        was just drawn by the player - or if it is not the player's
        turn.
        """
        result = []

        for card in ring.get_player_hand(player):
            if card.color != 'N' and card.color == carpet.current_color:
                result.append((CardAction(player, card),
                               Effects[card.type](player, card)))
        return result


class TypeRule(Rule):
    """Rule allowing to play a card with the same type."""
    def __init__(self):
        Rule.__init__(self)

    @no_referee
    @no_kitty
    @at_turn
    def check(self, player, ring, carpet):
        """Check this rule.

        This rule should match only for cards with type matching the
        heap top card type. It always fails if a kitty is present or
        was just drawn by the player - or if it is not the player's
        turn.
        """
        result = []

        for card in ring.get_player_hand(player):
            if card.color != 'N' and card.type == carpet.heap.peek().type:
                result.append((CardAction(player, card),
                               Effects[card.type](player, card)))
        return result


class DrawFourCardRule(Rule):
    """Rule enforcing the Draw Four card."""
    def __init__(self):
        Rule.__init__(self)

    @no_referee
    @no_kitty
    @at_turn
    def check(self, player, ring, carpet):
        """Check this rule.

        This rule should match only for Draw Four cards, whatever the
        heap. It always fails if there is a kitty or the kitty was
        just picked up - or if it is not the player's turn.
        """
        result = []

        effect = Effects['X']
        for card in ring.get_player_hand(player):
            if card.type == 'X':
                result.append((CardAction(player, card, 'Y'), effect(player, card, 'Y')))
                result.append((CardAction(player, card, 'G'), effect(player, card, 'G')))
                result.append((CardAction(player, card, 'R'), effect(player, card, 'R')))
                result.append((CardAction(player, card, 'B'), effect(player, card, 'B')))
        return result


class WildCardRule(Rule):
    """Rule enforcing the Wild card."""
    def __init__(self):
        Rule.__init__(self)

    @no_referee
    @no_kitty
    @at_turn
    def check(self, player, ring, carpet):
        """Check this rule.

        This rule should match only for Wild cards, whatever the
        heap. It always fails if a kitty is present or was just picked
        up - or if it is not the player's turn.
        """
        result = []

        effect = Effects['W']
        for card in ring.get_player_hand(player):
            if card.type == 'W':
                result.append((CardAction(player, card, 'Y'), effect(player, card, 'Y')))
                result.append((CardAction(player, card, 'G'), effect(player, card, 'G')))
                result.append((CardAction(player, card, 'R'), effect(player, card, 'R')))
                result.append((CardAction(player, card, 'B'), effect(player, card, 'B')))
        return result


class DrawRule(Rule):
    """Rule enforcing the draw move."""
    def __init__(self):
        Rule.__init__(self)

    @no_referee
    @at_turn
    def check(self, player, ring, carpet):
        """Check this rule.
        """
        result = []

        if not ring.has_player_drawn(player):
            result.append((DrawAction(player), DrawEffect(player)))

        return result


class PassRule(Rule):
    """Rule enforcing the pass move."""
    def __init__(self):
        Rule.__init__(self)

    @no_referee
    @at_turn
    def check(self, player, ring, carpet):
        """Check this rule.

        A player is authorized to pass only if he has previously taken
        the kitty on penalty, or if he tried to pick up a card in
        other cases.
        """
        result = []

        if ring.has_player_drawn(player):
            result.append((PassAction(player), PassEffect(player)))

        return result


class FirstCardRule(Rule):
    """Rule used to check the first card returned by the referee."""
    def __init__(self):
        Rule.__init__(self)

    @only_referee
    def check(self, player, ring, carpet):
        """Check this rule.
        """
        result = []
        if not len(carpet.heap) and ring.is_ready():
            result.append((DrawAction(None), FirstCardEffect(None)))

        return result


class ReadyRule(Rule):
    """Rule used by players to notify the referee to return the first card."""
    def __init__(self):
        Rule.__init__(self)

    @no_referee
    def check(self, player, ring, carpet):
        """Check this rule.

        A player can be ready only at the beginning of the game, when
        he just gets his first hand and the first card is not returned
        on the heap.
        """
        result = []
        if not ring.is_ready(player):
            result.append((ReadyAction(player), ReadyEffect(player)))

        return result


class DealRule(Rule):
    """Rule used to start the game."""

    def __init__(self, cards):
        Rule.__init__(self)
        self._cards = cards

    @only_referee
    def check(self, player, ring, carpet):
        """Check this rule.
        """
        result = []

        # Check that the first player has no card should suffice here... no ?
        if not len(carpet.heap) and not len(ring[0].hand):
            result.append((None, DealEffect(None, self._cards)))

        return result


class EmptyDeckRule(Rule):
    """Rule used to handle an empty deck during the game."""

    def __init__(self):
        Rule.__init__(self)

    @only_referee
    def check(self, player, ring, carpet):
        """Check this rule.
        """
        result = []

        # Check that the first player has no card should suffice here... no ?
        if len(carpet.heap) and not len(carpet.deck):
            result.append((None, SwapStacksEffect()))

        return result


class VictoryRule(Rule):
    """Rule used to handle the end of the game."""

    def __init__(self):
        Rule.__init__(self)

    @only_referee
    def check(self, player, ring, carpet):
        """Check this rule.

        Victory is pronounced if one player does not have anymore
        cards.
        """
        result = []

        if len(carpet.heap):
            for player in ring.get_list():
                if ring.get_player_hand(player) == []:
                    result.append((None, EndOfGameEffect(player)))
                    break

        return result


Effects = {
    'D': DrawTwoCardEffect,
    'P': PassCardEffect,
    'R': ReverseCardEffect,
    'W': CardEffect,
    'X': DrawFourCardEffect }
for i in range(10):
    Effects[str(i)] = CardEffect


def init(options):
    rules = set()
    rules.add(ColorRule())
    rules.add(DealRule(options['cards']))
    rules.add(DrawRule())
    rules.add(DrawFourCardRule())
    rules.add(FirstCardRule())
    rules.add(PassRule())
    rules.add(ReadyRule())
    rules.add(TypeRule())
    rules.add(WildCardRule())
    rules.add(EmptyDeckRule())
    rules.add(VictoryRule())

    return rules
